import greenfoot.*;

/**
 * The DNA for a creature. Contains things such as speed and how much to turn in certain curcumstances.
 * 
 * @author TriMill
 * @version 1.0.1
 */
public class CreatureDNA  
{
    /**
     * Speed of the creature
     */
    private double speed;
    /**
     * How much the creature turns on odd clock cycles
     */
    private int oddTurn;
    /**
     * How much the creature turns on even clock cycles
     */
    private int evenTurn;
    /**
     * How much the creature turns when it hits food
     */
    private int foodTurn;
    /**
     * How fit the creture is, determined by size, speed, and fullness.
     */
    private double fitness;
    /**
     * What direction the creature faces when it hatches.
     */
    private int initTurn;
    /**
     * How big the creature is
     */
    private int size;
    
    
    /**
     * How many mutations have occured during the simulation
     */
    private static int runningMutations;
    /**
     * How many reproductions have occured during the simulation
     */
    private static int runningReproductions;
    // Initialize them with 0.
    static
    {
        runningMutations = 0;
        runningReproductions = 0;
    }
    
    /**
     * Creates new, random creature DNA.
     */
    public CreatureDNA(double startingFood)
    {
        speed = (Greenfoot.getRandomNumber(8) + 2)/(Greenfoot.getRandomNumber(4) + 2.25) + 1;
        oddTurn = Greenfoot.getRandomNumber(181) - 90;
        evenTurn = Greenfoot.getRandomNumber(181) - 90;
        foodTurn = Greenfoot.getRandomNumber(361) - 180;
        initTurn = Greenfoot.getRandomNumber(181) - 90;
        size = Greenfoot.getRandomNumber(5) + 3;
        updateFitness(startingFood);
    }
    
    /**
     * Creates new creature DNA as DNA of offspring of two parent's DNA.
     * Randomly selects some genomes from one parent and some from the other, then applies some subtle mutations.
     */
    public CreatureDNA(CreatureDNA dna1, CreatureDNA dna2)
    {
        speed = choose(dna1.speed, dna2.speed);
        oddTurn = (int)choose(dna1.oddTurn, dna2.oddTurn);
        evenTurn = (int)choose(dna1.evenTurn, dna2.evenTurn);
        foodTurn = (int)choose(dna1.foodTurn, dna2.foodTurn);
        initTurn = (int)choose(dna1.initTurn, dna2.initTurn);
        size = (int)choose(dna1.size, dna2.size);
        
        speed = mutate(speed);
        oddTurn = (int)mutate(oddTurn);
        evenTurn = (int)mutate(evenTurn);
        foodTurn = (int)mutate(foodTurn);
        initTurn = (int)mutate(mutate(mutate(initTurn)));
        size = (int)mutate(size);
        
        runningReproductions++;
    }
    
    /**
     * Recalculates the fitness
     */
    public void updateFitness(double currentFood) 
    {
        fitness = speed/2 + currentFood * Math.round(speed/20.0 + 1.0) + size/1.5;
    }
    
    /**
     * Randomly chooses between two numbers
     */
    private double choose(double d1, double d2) 
    {
        if(Greenfoot.getRandomNumber(2)==0) 
            return d1;
        else
            return d2;
    }
    
    /**
     * Randomly increments or decrements a number
     */
    private double mutate(double number) 
    {
        double init = number;
        if(Greenfoot.getRandomNumber(250) == 0)
            number++;
        if(Greenfoot.getRandomNumber(250) == 0)
            number--;
        if(Greenfoot.getRandomNumber(750) == 0)
            number++;
        if(Greenfoot.getRandomNumber(750) == 0)
            number--;
        if(Greenfoot.getRandomNumber(2250) == 0)
            number++;
        if(Greenfoot.getRandomNumber(2250) == 0)
            number--;
        if(init != number)
            runningMutations++;
        return number;
    }
    
    /**
     * Get the speed of the Creature who has this DNA
     */
    public double getSpeed() { return speed; }
    /**
     * Get the turn on odd clock cycles of the Creature who has this DNA
     */
    public int getOddTurn() { return oddTurn; }
    /**
     * Get the turn on even clock cycles of the Creature who has this DNA
     */
    public int getEvenTurn() { return evenTurn; }
    /**
     * Get the turn on contact with food of the Creature who has this DNA
     */
    public int getFoodTurn() { return foodTurn; }
    /**
     * Get the initial rotation of the Creature who has this DNA
     */
    public int getInitialTurn() { return initTurn; }
    /**
     * Get the size of the Creature who has this DNA
     */
    public int getSize() { return size; }
    
    /**
     * Get the fitness value of the Creature who has this DNA
     */
    public double getFitness(double food) { updateFitness(food); return fitness; }
    
    /**
     * Get the number of total mutations in this simulation
     */
    public static double getMutations() { return runningMutations; }
    /**
     * Get the number of total reproductions in this simulation
     */
    public static double getReproductions() { return runningReproductions; }
}



